@@ -29,7 +29,6 @@ logging=init
 want_syslog=0
 syslog_tag=
 user='@MYSQLD_USER@'
-group='@MYSQLD_GROUP@'
 pid_file=
 err_log=
 err_log_base=
@@ -300,7 +299,6 @@ parse_arguments() {
       --pid[-_]file=*) pid_file="$val" ;;
       --plugin[-_]dir=*) PLUGIN_DIR="$val" ;;
       --user=*) user="$val"; SET_USER=1 ;;
-      --group=*) group="$val"; SET_USER=1 ;;
       --log[-_]basename=*|--hostname=*|--loose[-_]log[-_]basename=*)
         pid_file="$val.pid";
 	err_log_base="$val";
@@ -535,7 +533,8 @@ else
   ledir='@libexecdir@'
 fi
 
-helper=`find_in_bin mysqld_safe_helper`
+helper=`find_in_bin mariadbd-safe-helper`
+
 print_defaults=`find_in_bin my_print_defaults`
 # Check if helper exists
 command -v $helper --help >/dev/null 2>&1
@@ -703,8 +702,6 @@ then
   if test "$user" != "root" -o $SET_USER = 1
   then
     USER_OPTION="--user=$user"
-    # To be used if/when we enable --system-group as an option to mysqld
-    GROUP_OPTION="--group=$group"
   fi
   if test -n "$open_files"
   then
@@ -727,12 +724,7 @@ then
     log_error "Fatal error Can't create database directory '$mysql_unix_port'"
     exit 1
   fi
-  if [ "$user" -a "$group" ]; then
-    chown $user:$group $mysql_unix_port_dir
-  else
-    [ "$user" ] && chown $user $mysql_unix_port_dir
-    [ "$group" ] && chgrp $group $mysql_unix_port_dir
-  fi
+  chown $user $mysql_unix_port_dir
   chmod 755 $mysql_unix_port_dir
 fi
 
