@@ -24,7 +24,7 @@
 #include "item_cmpfunc.h"      // Item_bool_func
 #include "item_strfunc.h"      // Item_str_func
 #include "item_sum.h"
-#include "sql_type_json.h"
+
 
 class json_path_with_flags
 {
@@ -41,6 +41,13 @@ public:
 };
 
 
+void report_path_error_ex(const char *ps, json_path_t *p,
+                          const char *fname, int n_param,
+                          Sql_condition::enum_warning_level lv);
+void report_json_error_ex(const char *js, json_engine_t *je,
+                          const char *fname, int n_param,
+                          Sql_condition::enum_warning_level lv);
+
 class Json_engine_scan: public json_engine_t
 {
 public:
@@ -75,23 +82,28 @@ protected:
 
 public:
   Item_func_json_valid(THD *thd, Item *json) : Item_bool_func(thd, json) {}
-  longlong val_int();
-  const char *func_name() const { return "json_valid"; }
-  bool fix_length_and_dec()
+  longlong val_int() override;
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_valid") };
+    return name;
+  }
+  bool fix_length_and_dec() override
   {
     if (Item_bool_func::fix_length_and_dec())
       return TRUE;
-    maybe_null= 1;
+    set_maybe_null();
     return FALSE;
   }
   bool set_format_by_check_constraint(Send_field_extended_metadata *to) const
+    override
   {
     static const Lex_cstring fmt(STRING_WITH_LEN("json"));
     return to->set_format_name(fmt);
   }
-  Item *get_copy(THD *thd)
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_json_valid>(thd, this); }
-  enum Functype functype() const   { return JSON_VALID_FUNC; }
+  enum Functype functype() const override { return JSON_VALID_FUNC; }
 };
 
 
@@ -104,11 +116,15 @@ protected:
 public:
   Item_func_json_exists(THD *thd, Item *js, Item *i_path):
     Item_bool_func(thd, js, i_path) {}
-  const char *func_name() const { return "json_exists"; }
-  bool fix_length_and_dec();
-  Item *get_copy(THD *thd)
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_exists") };
+    return name;
+  }
+  bool fix_length_and_dec() override;
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_json_exists>(thd, this); }
-  longlong val_int();
+  longlong val_int() override;
 };
 
 
@@ -123,9 +139,12 @@ public:
    :Item_str_func(thd, a, b) { }
   Item_json_func(THD *thd, List<Item> &list)
    :Item_str_func(thd, list) { }
-  const Type_handler *type_handler() const override
+  bool is_json_type() override { return true; }
+  void make_send_field(THD *thd, Send_field *tmp_field) override
   {
-    return Type_handler_json_common::json_type_handler(max_length);
+    Item_str_func::make_send_field(thd, tmp_field);
+    static const Lex_cstring fmt(STRING_WITH_LEN("json"));
+    tmp_field->set_format_name(fmt);
   }
 };
 
@@ -137,7 +156,11 @@ class Item_func_json_value: public Item_str_func,
 public:
   Item_func_json_value(THD *thd, Item *js, Item *i_path):
     Item_str_func(thd, js, i_path) {}
-  const char *func_name() const override { return "json_value"; }
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_value") };
+    return name;
+  }
   bool fix_length_and_dec() override ;
   String *val_str(String *to) override
   {
@@ -161,7 +184,11 @@ class Item_func_json_query: public Item_json_func,
 public:
   Item_func_json_query(THD *thd, Item *js, Item *i_path):
     Item_json_func(thd, js, i_path) {}
-  const char *func_name() const override { return "json_query"; } 
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_query") };
+    return name;
+  }
   bool fix_length_and_dec() override;
   String *val_str(String *to) override
   {
@@ -186,10 +213,14 @@ protected:
 
 public:
   Item_func_json_quote(THD *thd, Item *s): Item_str_func(thd, s) {}
-  const char *func_name() const { return "json_quote"; }
-  bool fix_length_and_dec();
-  String *val_str(String *);
-  Item *get_copy(THD *thd)
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_quote") };
+    return name;
+  }
+  bool fix_length_and_dec() override;
+  String *val_str(String *) override;
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_json_quote>(thd, this); }
 };
 
@@ -201,10 +232,14 @@ protected:
   String *read_json(json_engine_t *je);
 public:
   Item_func_json_unquote(THD *thd, Item *s): Item_str_func(thd, s) {}
-  const char *func_name() const { return "json_unquote"; }
-  bool fix_length_and_dec();
-  String *val_str(String *);
-  Item *get_copy(THD *thd)
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_unquote") };
+    return name;
+  }
+  bool fix_length_and_dec() override;
+  String *val_str(String *) override;
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_json_unquote>(thd, this); }
 };
 
@@ -232,15 +267,19 @@ public:
                     char **out_val, int *value_len);
   Item_func_json_extract(THD *thd, List<Item> &list):
     Item_json_str_multipath(thd, list) {}
-  const char *func_name() const { return "json_extract"; }
-  enum Functype functype() const   { return JSON_EXTRACT_FUNC; }
-  bool fix_length_and_dec();
-  String *val_str(String *);
-  longlong val_int();
-  double val_real();
-  my_decimal *val_decimal(my_decimal *);
-  uint get_n_paths() const { return arg_count - 1; }
-  Item *get_copy(THD *thd)
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_extract") };
+    return name;
+  }
+  enum Functype functype() const override { return JSON_EXTRACT_FUNC; }
+  bool fix_length_and_dec() override;
+  String *val_str(String *) override;
+  longlong val_int() override;
+  double val_real() override;
+  my_decimal *val_decimal(my_decimal *) override;
+  uint get_n_paths() const override { return arg_count - 1; }
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_json_extract>(thd, this); }
 };
 
@@ -256,10 +295,14 @@ protected:
 public:
   Item_func_json_contains(THD *thd, List<Item> &list):
     Item_bool_func(thd, list) {}
-  const char *func_name() const { return "json_contains"; }
-  bool fix_length_and_dec();
-  longlong val_int();
-  Item *get_copy(THD *thd)
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_contains") };
+    return name;
+  }
+  bool fix_length_and_dec() override;
+  longlong val_int() override;
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_json_contains>(thd, this); }
 };
 
@@ -277,12 +320,16 @@ protected:
 public:
   Item_func_json_contains_path(THD *thd, List<Item> &list):
     Item_bool_func(thd, list), tmp_paths(0) {}
-  const char *func_name() const { return "json_contains_path"; }
-  bool fix_fields(THD *thd, Item **ref);
-  bool fix_length_and_dec();
-  void cleanup();
-  longlong val_int();
-  Item *get_copy(THD *thd)
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_contains_path") };
+    return name;
+  }
+  bool fix_fields(THD *thd, Item **ref) override;
+  bool fix_length_and_dec() override;
+  void cleanup() override;
+  longlong val_int() override;
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_json_contains_path>(thd, this); }
 };
 
@@ -297,10 +344,14 @@ public:
     Item_json_func(thd) {}
   Item_func_json_array(THD *thd, List<Item> &list):
     Item_json_func(thd, list) {}
-  String *val_str(String *);
-  bool fix_length_and_dec();
-  const char *func_name() const { return "json_array"; }
-  Item *get_copy(THD *thd)
+  String *val_str(String *) override;
+  bool fix_length_and_dec() override;
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_array") };
+    return name;
+  }
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_json_array>(thd, this); }
 };
 
@@ -313,11 +364,15 @@ protected:
 public:
   Item_func_json_array_append(THD *thd, List<Item> &list):
     Item_json_str_multipath(thd, list) {}
-  bool fix_length_and_dec();
-  String *val_str(String *);
-  uint get_n_paths() const { return arg_count/2; }
-  const char *func_name() const { return "json_array_append"; }
-  Item *get_copy(THD *thd)
+  bool fix_length_and_dec() override;
+  String *val_str(String *) override;
+  uint get_n_paths() const override { return arg_count/2; }
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_array_append") };
+    return name;
+  }
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_json_array_append>(thd, this); }
 };
 
@@ -327,9 +382,13 @@ class Item_func_json_array_insert: public Item_func_json_array_append
 public:
   Item_func_json_array_insert(THD *thd, List<Item> &list):
     Item_func_json_array_append(thd, list) {}
-  String *val_str(String *);
-  const char *func_name() const { return "json_array_insert"; }
-  Item *get_copy(THD *thd)
+  String *val_str(String *) override;
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_array_insert") };
+    return name;
+  }
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_json_array_insert>(thd, this); }
 };
 
@@ -341,9 +400,13 @@ public:
     Item_func_json_array(thd) {}
   Item_func_json_object(THD *thd, List<Item> &list):
     Item_func_json_array(thd, list) {}
-  String *val_str(String *);
-  const char *func_name() const { return "json_object"; }
-  Item *get_copy(THD *thd)
+  String *val_str(String *) override;
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_object") };
+    return name;
+  }
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_json_object>(thd, this); }
 };
 
@@ -355,9 +418,13 @@ protected:
 public:
   Item_func_json_merge(THD *thd, List<Item> &list):
     Item_func_json_array(thd, list) {}
-  String *val_str(String *);
-  const char *func_name() const { return "json_merge_preserve"; }
-  Item *get_copy(THD *thd)
+  String *val_str(String *) override;
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_merge_preserve") };
+    return name;
+  }
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_json_merge>(thd, this); }
 };
 
@@ -366,19 +433,23 @@ class Item_func_json_merge_patch: public Item_func_json_merge
 public:
   Item_func_json_merge_patch(THD *thd, List<Item> &list):
     Item_func_json_merge(thd, list) {}
-  const char *func_name() const { return "json_merge_patch"; }
-  String *val_str(String *);
-  Item *get_copy(THD *thd)
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_merge_patch") };
+    return name;
+  }
+  String *val_str(String *) override;
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_json_merge_patch>(thd, this); }
 };
 
 class Item_func_json_length: public Item_long_func
 {
-  bool check_arguments() const
+  bool check_arguments() const override
   {
-    return args[0]->check_type_can_return_text(func_name()) ||
+    return args[0]->check_type_can_return_text(func_name_cstring()) ||
            (arg_count > 1 &&
-            args[1]->check_type_general_purpose_string(func_name()));
+            args[1]->check_type_general_purpose_string(func_name_cstring()));
   }
 protected:
   json_path_with_flags path;
@@ -387,26 +458,34 @@ protected:
 public:
   Item_func_json_length(THD *thd, List<Item> &list):
     Item_long_func(thd, list) {}
-  const char *func_name() const { return "json_length"; }
-  bool fix_length_and_dec();
-  longlong val_int();
-  Item *get_copy(THD *thd)
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_length") };
+    return name;
+  }
+  bool fix_length_and_dec() override;
+  longlong val_int() override;
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_json_length>(thd, this); }
 };
 
 
 class Item_func_json_depth: public Item_long_func
 {
-  bool check_arguments() const
-  { return args[0]->check_type_can_return_text(func_name()); }
+  bool check_arguments() const override
+  { return args[0]->check_type_can_return_text(func_name_cstring()); }
 protected:
   String tmp_js;
 public:
   Item_func_json_depth(THD *thd, Item *js): Item_long_func(thd, js) {}
-  const char *func_name() const { return "json_depth"; }
-  bool fix_length_and_dec() { max_length= 10; return FALSE; }
-  longlong val_int();
-  Item *get_copy(THD *thd)
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_depth") };
+    return name;
+  }
+  bool fix_length_and_dec() override { max_length= 10; return FALSE; }
+  longlong val_int() override;
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_json_depth>(thd, this); }
 };
 
@@ -417,10 +496,14 @@ protected:
   String tmp_js;
 public:
   Item_func_json_type(THD *thd, Item *js): Item_str_func(thd, js) {}
-  const char *func_name() const { return "json_type"; }
-  bool fix_length_and_dec();
-  String *val_str(String *);
-  Item *get_copy(THD *thd)
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_type") };
+    return name;
+  }
+  bool fix_length_and_dec() override;
+  String *val_str(String *) override;
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_json_type>(thd, this); }
 };
 
@@ -435,15 +518,18 @@ public:
   Item_func_json_insert(bool i_mode, bool r_mode, THD *thd, List<Item> &list):
     Item_json_str_multipath(thd, list),
       mode_insert(i_mode), mode_replace(r_mode) {}
-  bool fix_length_and_dec();
-  String *val_str(String *);
-  uint get_n_paths() const { return arg_count/2; }
-  const char *func_name() const
+  bool fix_length_and_dec() override;
+  String *val_str(String *) override;
+  uint get_n_paths() const override { return arg_count/2; }
+  LEX_CSTRING func_name_cstring() const override
   {
-    return mode_insert ?
-             (mode_replace ? "json_set" : "json_insert") : "json_replace";
+    static LEX_CSTRING json_set=    {STRING_WITH_LEN("json_set") };
+    static LEX_CSTRING json_insert= {STRING_WITH_LEN("json_insert") };
+    static LEX_CSTRING json_replace= {STRING_WITH_LEN("json_replace") };
+    return (mode_insert ?
+            (mode_replace ? json_set : json_insert) : json_replace);
   }
-  Item *get_copy(THD *thd)
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_json_insert>(thd, this); }
 };
 
@@ -455,11 +541,15 @@ protected:
 public:
   Item_func_json_remove(THD *thd, List<Item> &list):
     Item_json_str_multipath(thd, list) {}
-  bool fix_length_and_dec();
-  String *val_str(String *);
-  uint get_n_paths() const { return arg_count - 1; }
-  const char *func_name() const { return "json_remove"; }
-  Item *get_copy(THD *thd)
+  bool fix_length_and_dec() override;
+  String *val_str(String *) override;
+  uint get_n_paths() const override { return arg_count - 1; }
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_remove") };
+    return name;
+  }
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_json_remove>(thd, this); }
 };
 
@@ -473,10 +563,14 @@ protected:
 public:
   Item_func_json_keys(THD *thd, List<Item> &list):
     Item_str_func(thd, list) {}
-  const char *func_name() const { return "json_keys"; }
-  bool fix_length_and_dec();
-  String *val_str(String *);
-  Item *get_copy(THD *thd)
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_keys") };
+    return name;
+  }
+  bool fix_length_and_dec() override;
+  String *val_str(String *) override;
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_json_keys>(thd, this); }
 };
 
@@ -496,12 +590,16 @@ protected:
 public:
   Item_func_json_search(THD *thd, List<Item> &list):
     Item_json_str_multipath(thd, list) {}
-  const char *func_name() const { return "json_search"; }
-  bool fix_fields(THD *thd, Item **ref);
-  bool fix_length_and_dec();
-  String *val_str(String *);
-  uint get_n_paths() const { return arg_count > 4 ? arg_count - 4 : 0; }
-  Item *get_copy(THD *thd)
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_search") };
+    return name;
+  }
+  bool fix_fields(THD *thd, Item **ref) override;
+  bool fix_length_and_dec() override;
+  String *val_str(String *) override;
+  uint get_n_paths() const override { return arg_count > 4 ? arg_count - 4 : 0; }
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_json_search>(thd, this); }
 };
 
@@ -525,11 +623,11 @@ public:
   Item_func_json_format(THD *thd, List<Item> &list):
     Item_json_func(thd, list), fmt(DETAILED) {}
 
-  const char *func_name() const;
-  bool fix_length_and_dec();
-  String *val_str(String *str);
-  String *val_json(String *str);
-  Item *get_copy(THD *thd)
+  LEX_CSTRING func_name_cstring() const override;
+  bool fix_length_and_dec() override;
+  String *val_str(String *str) override;
+  String *val_json(String *str) override;
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_json_format>(thd, this); }
 };
 
@@ -559,12 +657,13 @@ public:
   }
   Item_func_json_arrayagg(THD *thd, Item_func_json_arrayagg *item) :
     Item_func_group_concat(thd, item) {}
-  const Type_handler *type_handler() const override
-  {
-    return Type_handler_json_common::json_type_handler_sum(this);
-  }
+  bool is_json_type() override { return true; }
 
-  const char *func_name() const override { return "json_arrayagg("; }
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_arrayagg(") };
+    return name;
+  }
   enum Sumfunctype sum_func() const override { return JSON_ARRAYAGG_FUNC; }
 
   String* val_str(String *str) override;
@@ -583,17 +682,24 @@ public:
     Item_sum(thd, key, value)
   {
     quick_group= FALSE;
-    result.append("{");
+    result.append('{');
   }
 
   Item_func_json_objectagg(THD *thd, Item_func_json_objectagg *item);
+  bool is_json_type() override { return true; }
   void cleanup() override;
 
-  enum Sumfunctype sum_func() const override {return JSON_OBJECTAGG_FUNC;}
-  const char *func_name() const override { return "json_objectagg"; }
+  enum Sumfunctype sum_func () const override { return JSON_OBJECTAGG_FUNC;}
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_objectagg") };
+    return name;
+  }
   const Type_handler *type_handler() const override
   {
-    return Type_handler_json_common::json_type_handler_sum(this);
+    if (too_big_for_varchar())
+      return &type_handler_blob;
+    return &type_handler_varchar;
   }
   void clear() override;
   bool add() override;
