@@ -26,7 +26,7 @@
   Functions for easy reading of records, possible through a cache
 */
 
-#include <my_global.h>
+#include "mariadb.h"
 #include "records.h"
 #include "sql_priv.h"
 #include "records.h"
@@ -38,8 +38,8 @@
 static int rr_quick(READ_RECORD *info);
 int rr_sequential(READ_RECORD *info);
 static int rr_from_tempfile(READ_RECORD *info);
-static int rr_unpack_from_tempfile(READ_RECORD *info);
-static int rr_unpack_from_buffer(READ_RECORD *info);
+template<bool> static int rr_unpack_from_tempfile(READ_RECORD *info);
+template<bool,bool> static int rr_unpack_from_buffer(READ_RECORD *info);
 int rr_from_pointers(READ_RECORD *info);
 static int rr_from_cache(READ_RECORD *info);
 static int init_rr_cache(THD *thd, READ_RECORD *info);
@@ -77,20 +77,19 @@ bool init_read_record_idx(READ_RECORD *info, THD *thd, TABLE *table,
   bzero((char*) info,sizeof(*info));
   info->thd= thd;
   info->table= table;
-  info->record= table->record[0];
   info->print_error= print_error;
   info->unlock_row= rr_unlock_row;
 
   table->status=0;			/* And it's always found */
   if (!table->file->inited &&
-      (error= table->file->ha_index_init(idx, 1)))
+      unlikely(error= table->file->ha_index_init(idx, 1)))
   {
     if (print_error)
       table->file->print_error(error, MYF(0));
   }
 
-  /* read_record will be changed to rr_index in rr_index_first */
-  info->read_record= reverse ? rr_index_last : rr_index_first;
+  /* read_record_func will be changed to rr_index in rr_index_first */
+  info->read_record_func= reverse ? rr_index_last : rr_index_first;
   DBUG_RETURN(error != 0);
 }
 
@@ -188,35 +187,33 @@ bool init_read_record(READ_RECORD *info,THD *thd, TABLE *table,
                       bool disable_rr_cache)
 {
   IO_CACHE *tempfile;
-  SORT_ADDON_FIELD *addon_field= filesort ? filesort->addon_field : 0;
   DBUG_ENTER("init_read_record");
 
+  const bool using_addon_fields= filesort && filesort->using_addon_fields();
+  bool using_packed_sortkeys= filesort && filesort->using_packed_sortkeys();
+
   bzero((char*) info,sizeof(*info));
   info->thd=thd;
   info->table=table;
-  info->forms= &info->table;		/* Only one table */
-  info->addon_field= addon_field;
+  info->sort_info= filesort;
   
   if ((table->s->tmp_table == INTERNAL_TMP_TABLE) &&
-      !addon_field)
+      !using_addon_fields)
     (void) table->file->extra(HA_EXTRA_MMAP);
   
-  if (addon_field)
+  if (using_addon_fields)
   {
-    info->rec_buf=    (uchar*) filesort->addon_buf.str;
-    info->ref_length= filesort->addon_buf.length;
-    info->unpack=     filesort->unpack;
+    info->rec_buf=    filesort->addon_fields->get_addon_buf();
+    info->ref_length= filesort->addon_fields->get_addon_buf_length();
   }
   else
   {
     empty_record(table);
-    info->record= table->record[0];
-    info->ref_length= table->file->ref_length;
+    info->ref_length= (uint)table->file->ref_length;
   }
   info->select=select;
   info->print_error=print_error;
   info->unlock_row= rr_unlock_row;
-  info->ignore_not_found_rows= 0;
   table->status= 0;			/* Rows are always found */
 
   tempfile= 0;
@@ -227,14 +224,25 @@ bool init_read_record(READ_RECORD *info,THD *thd, TABLE *table,
 
   if (tempfile && !(select && select->quick))
   {
-    DBUG_PRINT("info",("using rr_from_tempfile"));
-    info->read_record= (addon_field ?
-                        rr_unpack_from_tempfile : rr_from_tempfile);
+    if (using_addon_fields)
+    {
+      DBUG_PRINT("info",("using rr_from_tempfile"));
+      if (filesort->addon_fields->using_packed_addons())
+        info->read_record_func= rr_unpack_from_tempfile<true>;
+      else
+        info->read_record_func= rr_unpack_from_tempfile<false>;
+    }
+    else
+    {
+      DBUG_PRINT("info",("using rr_from_tempfile"));
+      info->read_record_func= rr_from_tempfile;
+    }
+
     info->io_cache= tempfile;
     reinit_io_cache(info->io_cache,READ_CACHE,0L,0,0);
     info->ref_pos=table->file->ref;
     if (!table->file->inited)
-      if (table->file->ha_rnd_init_with_error(0))
+      if (unlikely(table->file->ha_rnd_init_with_error(0)))
         DBUG_RETURN(1);
 
     /*
@@ -243,11 +251,11 @@ bool init_read_record(READ_RECORD *info,THD *thd, TABLE *table,
       and filesort->io_cache is read sequentially
     */
     if (!disable_rr_cache &&
-        !addon_field &&
+        !using_addon_fields &&
 	thd->variables.read_rnd_buff_size &&
 	!(table->file->ha_table_flags() & HA_FAST_KEY_READ) &&
 	(table->db_stat & HA_READ_ONLY ||
-	 table->reginfo.lock_type <= TL_READ_NO_INSERT) &&
+	 table->reginfo.lock_type < TL_FIRST_WRITE) &&
 	(ulonglong) table->s->reclength* (table->file->stats.records+
                                           table->file->stats.deleted) >
 	(ulonglong) MIN_FILE_LENGTH_TO_USE_ROW_CACHE &&
@@ -259,32 +267,54 @@ bool init_read_record(READ_RECORD *info,THD *thd, TABLE *table,
       if (! init_rr_cache(thd, info))
       {
 	DBUG_PRINT("info",("using rr_from_cache"));
-	info->read_record=rr_from_cache;
+        info->read_record_func= rr_from_cache;
       }
     }
   }
   else if (select && select->quick)
   {
     DBUG_PRINT("info",("using rr_quick"));
-    info->read_record=rr_quick;
+    info->read_record_func= rr_quick;
   }
-  else if (filesort && filesort->record_pointers)
+  else if (filesort && filesort->has_filesort_result_in_memory())
   {
     DBUG_PRINT("info",("using record_pointers"));
-    if (table->file->ha_rnd_init_with_error(0))
+    if (unlikely(table->file->ha_rnd_init_with_error(0)))
       DBUG_RETURN(1);
+
     info->cache_pos= filesort->record_pointers;
-    info->cache_end= (info->cache_pos+ 
-                      filesort->return_rows * info->ref_length);
-    info->read_record= (addon_field ?
-                        rr_unpack_from_buffer : rr_from_pointers);
+    if (using_addon_fields)
+    {
+      DBUG_PRINT("info",("using rr_unpack_from_buffer"));
+      DBUG_ASSERT(filesort->sorted_result_in_fsbuf);
+      info->unpack_counter= 0;
+
+      if (filesort->using_packed_addons())
+      {
+        info->read_record_func= using_packed_sortkeys ?
+                                rr_unpack_from_buffer<true, true> :
+                                rr_unpack_from_buffer<true, false>;
+      }
+      else
+      {
+        info->read_record_func= using_packed_sortkeys ?
+                                rr_unpack_from_buffer<false, true> :
+                                rr_unpack_from_buffer<false, false>;
+      }
+    }
+    else
+    {
+      info->cache_end= (info->cache_pos+
+                        filesort->return_rows * info->ref_length);
+      info->read_record_func= rr_from_pointers;
+    }
   }
   else if (table->file->keyread_enabled())
   {
     int error;
-    info->read_record= rr_index_first;
+    info->read_record_func= rr_index_first;
     if (!table->file->inited &&
-        (error= table->file->ha_index_init(table->file->keyread, 1)))
+        unlikely((error= table->file->ha_index_init(table->file->keyread, 1))))
     {
       if (print_error)
         table->file->print_error(error, MYF(0));
@@ -294,8 +324,8 @@ bool init_read_record(READ_RECORD *info,THD *thd, TABLE *table,
   else
   {
     DBUG_PRINT("info",("using rr_sequential"));
-    info->read_record=rr_sequential;
-    if (table->file->ha_rnd_init_with_error(1))
+    info->read_record_func= rr_sequential;
+    if (unlikely(table->file->ha_rnd_init_with_error(1)))
       DBUG_RETURN(1);
     /* We can use record cache if we don't update dynamic length tables */
     if (!table->no_cache &&
@@ -325,9 +355,9 @@ void end_read_record(READ_RECORD *info)
   free_cache(info);
   if (info->table)
   {
-    if (info->table->db_stat) // if opened
+    if (info->table->is_created())
       (void) info->table->file->extra(HA_EXTRA_NO_CACHE);
-    if (info->read_record != rr_quick) // otherwise quick_range does it
+    if (info->read_record_func != rr_quick) // otherwise quick_range does it
       (void) info->table->file->ha_index_or_rnd_end();
     info->table=0;
   }
@@ -372,11 +402,8 @@ static int rr_quick(READ_RECORD *info)
   int tmp;
   while ((tmp= info->select->quick->get_next()))
   {
-    if (info->thd->killed || (tmp != HA_ERR_RECORD_DELETED))
-    {
-      tmp= rr_handle_error(info, tmp);
-      break;
-    }
+    tmp= rr_handle_error(info, tmp);
+    break;
   }
   return tmp;
 }
@@ -405,8 +432,8 @@ static int rr_index_first(READ_RECORD *info)
     return tmp;
   }
 
-  tmp= info->table->file->ha_index_first(info->record);
-  info->read_record= rr_index;
+  tmp= info->table->file->ha_index_first(info->record());
+  info->read_record_func= rr_index;
   if (tmp)
     tmp= rr_handle_error(info, tmp);
   return tmp;
@@ -428,8 +455,8 @@ static int rr_index_first(READ_RECORD *info)
 
 static int rr_index_last(READ_RECORD *info)
 {
-  int tmp= info->table->file->ha_index_last(info->record);
-  info->read_record= rr_index_desc;
+  int tmp= info->table->file->ha_index_last(info->record());
+  info->read_record_func= rr_index_desc;
   if (tmp)
     tmp= rr_handle_error(info, tmp);
   return tmp;
@@ -454,7 +481,7 @@ static int rr_index_last(READ_RECORD *info)
 
 static int rr_index(READ_RECORD *info)
 {
-  int tmp= info->table->file->ha_index_next(info->record);
+  int tmp= info->table->file->ha_index_next(info->record());
   if (tmp)
     tmp= rr_handle_error(info, tmp);
   return tmp;
@@ -479,7 +506,7 @@ static int rr_index(READ_RECORD *info)
 
 static int rr_index_desc(READ_RECORD *info)
 {
-  int tmp= info->table->file->ha_index_prev(info->record);
+  int tmp= info->table->file->ha_index_prev(info->record());
   if (tmp)
     tmp= rr_handle_error(info, tmp);
   return tmp;
@@ -489,17 +516,10 @@ static int rr_index_desc(READ_RECORD *info)
 int rr_sequential(READ_RECORD *info)
 {
   int tmp;
-  while ((tmp= info->table->file->ha_rnd_next(info->record)))
+  while ((tmp= info->table->file->ha_rnd_next(info->record())))
   {
-    /*
-      rnd_next can return RECORD_DELETED for MyISAM when one thread is
-      reading and another deleting without locks.
-    */
-    if (info->thd->killed || (tmp != HA_ERR_RECORD_DELETED))
-    {
-      tmp= rr_handle_error(info, tmp);
-      break;
-    }
+    tmp= rr_handle_error(info, tmp);
+    break;
   }
   return tmp;
 }
@@ -512,11 +532,10 @@ static int rr_from_tempfile(READ_RECORD *info)
   {
     if (my_b_read(info->io_cache,info->ref_pos,info->ref_length))
       return -1;					/* End of file */
-    if (!(tmp= info->table->file->ha_rnd_pos(info->record,info->ref_pos)))
+    if (!(tmp= info->table->file->ha_rnd_pos(info->record(), info->ref_pos)))
       break;
     /* The following is extremely unlikely to happen */
-    if (tmp == HA_ERR_RECORD_DELETED ||
-        (tmp == HA_ERR_KEY_NOT_FOUND && info->ignore_not_found_rows))
+    if (tmp == HA_ERR_KEY_NOT_FOUND)
       continue;
     tmp= rr_handle_error(info, tmp);
     break;
@@ -533,7 +552,11 @@ static int rr_from_tempfile(READ_RECORD *info)
   the fields values use in the result set from this buffer into their
   positions in the regular record buffer.
 
-  @param info          Reference to the context including record descriptors
+  @param info                 Reference to the context including record
+                              descriptors
+  @param Packed_addon_fields  Are the addon fields packed?
+                              This is a compile-time constant, to
+                              avoid if (....) tests during execution.
 
   @retval
     0   Record successfully read.
@@ -541,12 +564,38 @@ static int rr_from_tempfile(READ_RECORD *info)
     -1   There is no record to be read anymore.
 */
 
+template<bool Packed_addon_fields>
 static int rr_unpack_from_tempfile(READ_RECORD *info)
 {
-  if (my_b_read(info->io_cache, info->rec_buf, info->ref_length))
-    return -1;
-  (*info->unpack)(info->addon_field, info->rec_buf,
-                  info->rec_buf + info->ref_length);
+  uchar *destination= info->rec_buf;
+#ifndef DBUG_OFF
+  my_off_t where= my_b_tell(info->io_cache);
+#endif
+  if (Packed_addon_fields)
+  {
+    const uint len_sz= Addon_fields::size_of_length_field;
+
+    // First read length of the record.
+    if (my_b_read(info->io_cache, destination, len_sz))
+      return -1;
+    uint res_length= Addon_fields::read_addon_length(destination);
+    DBUG_PRINT("info", ("rr_unpack from %llu to %p sz %u",
+                        static_cast<ulonglong>(where),
+                        destination, res_length));
+    DBUG_ASSERT(res_length > len_sz);
+    DBUG_ASSERT(info->sort_info->using_addon_fields());
+
+    // Then read the rest of the record.
+    if (my_b_read(info->io_cache, destination + len_sz, res_length - len_sz))
+      return -1;                                /* purecov: inspected */
+  }
+  else
+  {
+    if (my_b_read(info->io_cache, destination, info->ref_length))
+      return -1;
+  }
+
+  info->sort_info->unpack_addon_fields<Packed_addon_fields>(destination);
 
   return 0;
 }
@@ -563,12 +612,11 @@ int rr_from_pointers(READ_RECORD *info)
     cache_pos= info->cache_pos;
     info->cache_pos+= info->ref_length;
 
-    if (!(tmp= info->table->file->ha_rnd_pos(info->record,cache_pos)))
+    if (!(tmp= info->table->file->ha_rnd_pos(info->record(), cache_pos)))
       break;
 
     /* The following is extremely unlikely to happen */
-    if (tmp == HA_ERR_RECORD_DELETED ||
-        (tmp == HA_ERR_KEY_NOT_FOUND && info->ignore_not_found_rows))
+    if (tmp == HA_ERR_KEY_NOT_FOUND)
       continue;
     tmp= rr_handle_error(info, tmp);
     break;
@@ -584,7 +632,11 @@ int rr_from_pointers(READ_RECORD *info)
   the fields values use in the result set from this buffer into their
   positions in the regular record buffer.
 
-  @param info          Reference to the context including record descriptors
+  @param info                 Reference to the context including record
+                              descriptors
+  @param Packed_addon_fields  Are the addon fields packed?
+                              This is a compile-time constant, to
+                              avoid if (....) tests during execution.
 
   @retval
     0   Record successfully read.
@@ -592,44 +644,54 @@ int rr_from_pointers(READ_RECORD *info)
     -1   There is no record to be read anymore.
 */
 
+template<bool Packed_addon_fields, bool Packed_sort_keys>
 static int rr_unpack_from_buffer(READ_RECORD *info)
 {
-  if (info->cache_pos == info->cache_end)
+  if (info->unpack_counter == info->sort_info->return_rows)
     return -1;                      /* End of buffer */
-  (*info->unpack)(info->addon_field, info->cache_pos,
-                  info->cache_end);
-  info->cache_pos+= info->ref_length;
+
+  uchar *record= info->sort_info->get_sorted_record(
+    static_cast<uint>(info->unpack_counter));
+
+  uint sort_length= Packed_sort_keys ?
+                    Sort_keys::read_sortkey_length(record):
+                    info->sort_info->get_sort_length();
+
+  uchar *plen= record + sort_length;
+  info->sort_info->unpack_addon_fields<Packed_addon_fields>(plen);
+  info->unpack_counter++;
   return 0;
 }
 	/* cacheing of records from a database */
 
+static const uint STRUCT_LENGTH= 3 + MAX_REFLENGTH;
+
 static int init_rr_cache(THD *thd, READ_RECORD *info)
 {
-  uint rec_cache_size;
+  uint rec_cache_size, cache_records;
   DBUG_ENTER("init_rr_cache");
 
-  info->struct_length= 3+MAX_REFLENGTH;
   info->reclength= ALIGN_SIZE(info->table->s->reclength+1);
-  if (info->reclength < info->struct_length)
-    info->reclength= ALIGN_SIZE(info->struct_length);
+  if (info->reclength < STRUCT_LENGTH)
+    info->reclength= ALIGN_SIZE(STRUCT_LENGTH);
 
   info->error_offset= info->table->s->reclength;
-  info->cache_records= (thd->variables.read_rnd_buff_size /
-                        (info->reclength+info->struct_length));
-  rec_cache_size= info->cache_records*info->reclength;
-  info->rec_cache_size= info->cache_records*info->ref_length;
+  cache_records= thd->variables.read_rnd_buff_size /
+                 (info->reclength + STRUCT_LENGTH);
+  rec_cache_size= cache_records * info->reclength;
+  info->rec_cache_size= cache_records * info->ref_length;
 
   // We have to allocate one more byte to use uint3korr (see comments for it)
-  if (info->cache_records <= 2 ||
-      !(info->cache=(uchar*) my_malloc_lock(rec_cache_size+info->cache_records*
-					   info->struct_length+1,
-					   MYF(MY_THREAD_SPECIFIC))))
+  if (cache_records <= 2 ||
+      !(info->cache= (uchar*) my_malloc_lock(rec_cache_size + cache_records *
+                                             STRUCT_LENGTH + 1,
+                                             MYF(MY_THREAD_SPECIFIC))))
     DBUG_RETURN(1);
 #ifdef HAVE_valgrind
   // Avoid warnings in qsort
-  bzero(info->cache,rec_cache_size+info->cache_records* info->struct_length+1);
+  bzero(info->cache, rec_cache_size + cache_records * STRUCT_LENGTH + 1);
 #endif
-  DBUG_PRINT("info",("Allocated buffert for %d records",info->cache_records));
+  DBUG_PRINT("info", ("Allocated buffer for %d records", cache_records));
   info->read_positions=info->cache+rec_cache_size;
   info->cache_pos=info->cache_end=info->cache;
   DBUG_RETURN(0);
@@ -649,7 +711,7 @@ static int rr_from_cache(READ_RECORD *info)
   {
     if (info->cache_pos != info->cache_end)
     {
-      if (info->cache_pos[info->error_offset])
+      if (unlikely(info->cache_pos[info->error_offset]))
       {
 	shortget(error,info->cache_pos);
 	if (info->print_error)
@@ -658,7 +720,7 @@ static int rr_from_cache(READ_RECORD *info)
       else
       {
 	error=0;
-	memcpy(info->record,info->cache_pos,
+        memcpy(info->record(), info->cache_pos,
                (size_t) info->table->s->reclength);
       }
       info->cache_pos+=info->reclength;
@@ -684,8 +746,7 @@ static int rr_from_cache(READ_RECORD *info)
       int3store(ref_position,(long) i);
       ref_position+=3;
     }
-    my_qsort(info->read_positions, length, info->struct_length,
-             (qsort_cmp) rr_cmp);
+    my_qsort(info->read_positions, length, STRUCT_LENGTH, (qsort_cmp) rr_cmp);
 
     position=info->read_positions;
     for (i=0 ; i < length ; i++)
@@ -695,7 +756,8 @@ static int rr_from_cache(READ_RECORD *info)
       record=uint3korr(position);
       position+=3;
       record_pos=info->cache+record*info->reclength;
-      if ((error=(int16) info->table->file->ha_rnd_pos(record_pos,info->ref_pos)))
+      if (unlikely((error= (int16) info->table->file->
+                    ha_rnd_pos(record_pos,info->ref_pos))))
       {
 	record_pos[info->error_offset]=1;
 	shortstore(record_pos,error);
@@ -732,3 +794,68 @@ static int rr_cmp(uchar *a,uchar *b)
   return (int) a[7] - (int) b[7];
 #endif
 }
+
+
+/**
+  Copy (unpack) values appended to sorted fields from a buffer back to
+  their regular positions specified by the Field::ptr pointers.
+
+  @param addon_field     Array of descriptors for appended fields
+  @param buff            Buffer which to unpack the value from
+
+  @note
+    The function is supposed to be used only as a callback function
+    when getting field values for the sorted result set.
+
+*/
+template<bool Packed_addon_fields>
+inline void SORT_INFO::unpack_addon_fields(uchar *buff)
+{
+  SORT_ADDON_FIELD *addonf= addon_fields->begin();
+  uchar *buff_end= buff + sort_buffer_size();
+  const uchar *start_of_record= buff + addonf->offset;
+
+  for ( ; addonf != addon_fields->end() ; addonf++)
+  {
+    Field *field= addonf->field;
+    if (addonf->null_bit && (addonf->null_bit & buff[addonf->null_offset]))
+    {
+      field->set_null();
+      continue;
+    }
+    field->set_notnull();
+    if (Packed_addon_fields)
+      start_of_record= field->unpack(field->ptr, start_of_record, buff_end, 0);
+    else
+      field->unpack(field->ptr, buff + addonf->offset, buff_end, 0);
+  }
+}
+
+
+/*
+  @brief
+    Read and unpack next record from a table
+
+  @details
+    The function first reads the next record from the table.
+    If a success then it unpacks the values to the base table fields.
+    This is used by SJM scan table to unpack the values of the materialized
+    table to the base table fields
+
+  @retval
+    0   Record successfully read.
+  @retval
+    -1   There is no record to be read anymore.
+    >0   Error
+*/
+int read_record_func_for_rr_and_unpack(READ_RECORD *info)
+{
+  int error;
+  if ((error= info->read_record_func_and_unpack_calls(info)))
+    return error;
+
+  for (Copy_field *cp= info->copy_field; cp != info->copy_field_end; cp++)
+    (*cp->do_copy)(cp);
+
+  return error;
+}
