@@ -144,8 +144,6 @@ int MBR::within(const MBR *mbr)
 
 /***************************** Gis_class_info *******************************/
 
-String Geometry::bad_geometry_data("Bad object", &my_charset_bin);
-
 Geometry::Class_info *Geometry::ci_collection[Geometry::wkb_last+1]=
 {
   NULL, NULL, NULL, NULL, NULL, NULL, NULL
@@ -241,9 +239,8 @@ Geometry::Class_info *Geometry::find_class(const char *name, size_t len)
   {
     if (*cur_rt &&
 	((*cur_rt)->m_name.length == len) &&
-	(my_strnncoll(&my_charset_latin1,
-		      (const uchar*) (*cur_rt)->m_name.str, len,
-		      (const uchar*) name, len) == 0))
+	(my_charset_latin1.strnncoll((*cur_rt)->m_name.str, len,
+		                     name, len) == 0))
       return *cur_rt;
   }
   return 0;
@@ -277,6 +274,33 @@ Geometry *Geometry::construct(Geometry_buffer *buffer,
 }
 
 
+uint Geometry::get_key_image_itMBR(LEX_CSTRING &src, uchar *buff, uint length)
+{
+  const char *dummy;
+  MBR mbr;
+  Geometry_buffer buffer;
+  Geometry *gobj;
+  const uint image_length= SIZEOF_STORED_DOUBLE*4;
+
+  if (src.length < SRID_SIZE)
+  {
+    bzero(buff, image_length);
+    return image_length;
+  }
+  gobj= Geometry::construct(&buffer, (char*) src.str, (uint32) src.length);
+  if (!gobj || gobj->get_mbr(&mbr, &dummy))
+    bzero(buff, image_length);
+  else
+  {
+    float8store(buff,    mbr.xmin);
+    float8store(buff+8,  mbr.xmax);
+    float8store(buff+16, mbr.ymin);
+    float8store(buff+24, mbr.ymax);
+  }
+  return image_length;
+}
+
+
 Geometry *Geometry::create_from_wkt(Geometry_buffer *buffer,
 				    Gis_read_stream *trs, String *wkt,
 				    bool init_stream)
@@ -356,7 +380,7 @@ int Geometry::as_json(String *wkt, uint max_dec_digits, const char **end)
   if (wkt->reserve(4 + type_keyname_len + 2 + len + 2 + 2 +
                    coord_keyname_len + 4, 512))
     return 1;
-  wkt->qs_append("\"", 1);
+  wkt->qs_append('"');
   wkt->qs_append((const char *) type_keyname, type_keyname_len);
   wkt->qs_append("\": \"", 4);
   wkt->qs_append(get_class_info()->m_geojson_name.str, len);
@@ -380,7 +404,7 @@ int Geometry::bbox_as_json(String *wkt)
   const char *end;
   if (wkt->reserve(5 + bbox_keyname_len + (FLOATING_POINT_DECIMALS+2)*4, 512))
     return 1;
-  wkt->qs_append("\"", 1);
+  wkt->qs_append('"');
   wkt->qs_append((const char *) bbox_keyname, bbox_keyname_len);
   wkt->qs_append("\": [", 4);
 
@@ -394,7 +418,7 @@ int Geometry::bbox_as_json(String *wkt)
   wkt->qs_append(mbr.xmax);
   wkt->qs_append(", ", 2);
   wkt->qs_append(mbr.ymax);
-  wkt->qs_append("]", 1);
+  wkt->qs_append(']');
 
   return 0;
 }
@@ -520,8 +544,8 @@ Geometry *Geometry::create_from_json(Geometry_buffer *buffer,
             goto create_geom;
         }
         else if (je->value_len == feature_coll_type_len &&
-            my_strnncoll(&my_charset_latin1, je->value, je->value_len,
-		         feature_coll_type, feature_coll_type_len) == 0)
+            my_charset_latin1.strnncoll(je->value, je->value_len,
+		                        feature_coll_type, feature_coll_type_len) == 0)
         {
           /*
             'FeatureCollection' type found. Handle the 'Featurecollection'/'features'
@@ -532,8 +556,8 @@ Geometry *Geometry::create_from_json(Geometry_buffer *buffer,
           fcoll_type_found= 1;
         }
         else if (je->value_len == feature_type_len &&
-                 my_strnncoll(&my_charset_latin1, je->value, je->value_len,
-		              feature_type, feature_type_len) == 0)
+                 my_charset_latin1.strnncoll(je->value, je->value_len,
+		                             feature_type, feature_type_len) == 0)
         {
           if (geometry_start)
             goto handle_geometry_key;
@@ -932,8 +956,7 @@ static int read_point_from_json(json_engine_t *je, bool er_on_3D,
       goto bad_coordinates;
 
     d= (n_coord == 0) ? x : ((n_coord == 1) ? y : &tmp);
-    *d= my_strntod(je->s.cs, (char *) je->value,
-                   je->value_len, &endptr, &err);
+    *d= je->s.cs->strntod((char *) je->value, je->value_len, &endptr, &err);
     if (err)
       goto bad_coordinates;
     n_coord++;
@@ -3257,9 +3280,7 @@ bool Gis_geometry_collection::init_from_wkt(Gis_read_stream *trs, String *wkb)
       return 1;
 
     if (next_word.length != 5 ||
-	(my_strnncoll(&my_charset_latin1,
-		      (const uchar*) "empty", 5,
-		      (const uchar*) next_word.str, 5) != 0))
+	(my_charset_latin1.strnncoll("empty", 5, next_word.str, 5) != 0))
     {
       for (;;)
       {
@@ -3489,13 +3510,13 @@ bool Gis_geometry_collection::get_data_as_json(String *txt, uint max_dec_digits,
     if (!(geom= create_by_typeid(&buffer, wkb_type)))
       return 1;
     geom->set_data_ptr(data, (uint) (m_data_end - data));
-    if (txt->append("{", 1) ||
+    if (txt->append('{') ||
         geom->as_json(txt, max_dec_digits, &data) ||
         txt->append(STRING_WITH_LEN("}, "), 512))
       return 1;
   }
   txt->length(txt->length() - 2);
-  if (txt->append("]", 1))
+  if (txt->append(']'))
     return 1;
 
   *end= data;
