@@ -31,14 +31,17 @@ ulong   wsrep_reject_queries;
 
 int wsrep_init_vars()
 {
-  wsrep_provider        = my_strdup(WSREP_NONE, MYF(MY_WME));
-  wsrep_provider_options= my_strdup("", MYF(MY_WME));
-  wsrep_cluster_address = my_strdup("", MYF(MY_WME));
-  wsrep_cluster_name    = my_strdup(WSREP_CLUSTER_NAME, MYF(MY_WME));
-  wsrep_node_name       = my_strdup("", MYF(MY_WME));
-  wsrep_node_address    = my_strdup("", MYF(MY_WME));
-  wsrep_node_incoming_address= my_strdup(WSREP_NODE_INCOMING_AUTO, MYF(MY_WME));
-  wsrep_start_position  = my_strdup(WSREP_START_POSITION_ZERO, MYF(MY_WME));
+  wsrep_provider        = my_strdup(PSI_INSTRUMENT_ME, WSREP_NONE, MYF(MY_WME));
+  wsrep_provider_options= my_strdup(PSI_INSTRUMENT_ME, "", MYF(MY_WME));
+  wsrep_cluster_address = my_strdup(PSI_INSTRUMENT_ME, "", MYF(MY_WME));
+  wsrep_cluster_name    = my_strdup(PSI_INSTRUMENT_ME, WSREP_CLUSTER_NAME, MYF(MY_WME));
+  wsrep_node_name       = my_strdup(PSI_INSTRUMENT_ME, "", MYF(MY_WME));
+  wsrep_node_address    = my_strdup(PSI_INSTRUMENT_ME, "", MYF(MY_WME));
+  wsrep_node_incoming_address= my_strdup(PSI_INSTRUMENT_ME, WSREP_NODE_INCOMING_AUTO, MYF(MY_WME));
+  if (wsrep_gtid_mode)
+    wsrep_start_position  = my_strdup(PSI_INSTRUMENT_ME, WSREP_START_POSITION_ZERO_GTID, MYF(MY_WME));
+  else
+    wsrep_start_position  = my_strdup(PSI_INSTRUMENT_ME, WSREP_START_POSITION_ZERO, MYF(MY_WME));
   return 0;
 }
 
@@ -48,7 +51,7 @@ static int get_provider_option_value(const char* opts,
 {
   int ret= 1;
   ulong opt_value_tmp;
-  char *opt_value_str, *s, *opts_copy= my_strdup(opts, MYF(MY_WME));
+  char *opt_value_str, *s, *opts_copy= my_strdup(PSI_INSTRUMENT_ME, opts, MYF(MY_WME));
 
   if ((opt_value_str= strstr(opts_copy, opt_name)) == NULL)
     goto end;
@@ -89,19 +92,15 @@ static bool refresh_provider_options()
   }
 }
 
-void wsrep_set_wsrep_on()
+void wsrep_set_wsrep_on(THD* thd)
 {
+  if (thd)
+    thd->wsrep_was_on= WSREP_ON_;
   WSREP_PROVIDER_EXISTS_= wsrep_provider &&
     strncasecmp(wsrep_provider, WSREP_NONE, FN_REFLEN);
   WSREP_ON_= global_system_variables.wsrep_on && WSREP_PROVIDER_EXISTS_;
 }
 
-/* This is intentionally declared as a weak global symbol, so that
-linking will succeed even if the server is built with a dynamically
-linked InnoDB. */
-ulong innodb_lock_schedule_algorithm __attribute__((weak));
-struct handlerton* innodb_hton_ptr __attribute__((weak));
-
 bool wsrep_on_update (sys_var *self, THD* thd, enum_var_type var_type)
 {
   if (var_type == OPT_GLOBAL)
@@ -131,7 +130,7 @@ bool wsrep_on_update (sys_var *self, THD* thd, enum_var_type var_type)
     thd->variables.wsrep_on= global_system_variables.wsrep_on= saved_wsrep_on;
   }
 
-  wsrep_set_wsrep_on();
+  wsrep_set_wsrep_on(thd);
 
   if (var_type == OPT_GLOBAL)
   {
@@ -155,14 +154,6 @@ bool wsrep_on_check(sys_var *self, THD* thd, set_var* var)
 
   if (new_wsrep_on)
   {
-    if (innodb_hton_ptr && innodb_lock_schedule_algorithm != 0)
-    {
-      my_message(ER_WRONG_ARGUMENTS, " WSREP (galera) can't be enabled "
-                 "if innodb_lock_schedule_algorithm=VATS. Please configure"
-                 " innodb_lock_schedule_algorithm=FCFS and restart.", MYF(0));
-      return true;
-    }
-
     if (!WSREP_PROVIDER_EXISTS)
     {
       my_message(ER_WRONG_ARGUMENTS, "WSREP (galera) can't be enabled "
@@ -237,6 +228,13 @@ bool wsrep_sync_wait_update (sys_var* self, THD* thd, enum_var_type var_type)
   return false;
 }
 
+template<typename T>
+static T parse_value(char** startptr, char** endptr)
+{
+   T val= strtoll(*startptr, *&endptr, 10);
+   *startptr= *endptr;
+   return val;
+}
 
 /*
   Verify the format of the given UUID:seqno.
@@ -270,12 +268,29 @@ bool wsrep_start_position_verify (const char* start_str)
     return true;
 
   char* endptr;
-  wsrep_seqno_t const seqno(strtoll(&start_str[uuid_len + 1], &endptr, 10));
+  char* startptr= (char *)start_str + uuid_len + 1;
+  wsrep_seqno_t const seqno(parse_value<uint64_t>(&startptr, &endptr));
 
   // Do not allow seqno < -1
-  if (*endptr == '\0' && seqno < -1)
+  if (seqno < -1)
     return true;
 
+  // Start parsing native GTID part
+  if (*startptr == ',')
+  {
+    startptr++;
+    uint32_t domain  __attribute__((unused))
+      (parse_value<uint32_t>(&startptr, &endptr));
+    if (*endptr != '-') return true;
+    startptr++;
+    uint32_t server  __attribute__((unused))
+      (parse_value<uint32_t>(&startptr, &endptr));
+    if (*endptr != '-') return true;
+    startptr++;
+    uint64_t seq  __attribute__((unused))
+      (parse_value<uint64_t>(&startptr, &endptr));
+  }
+
   // Remaining string was seqno.
   if (*endptr == '\0') return false;
 
@@ -287,9 +302,22 @@ static
 bool wsrep_set_local_position(THD* thd, const char* const value,
                               size_t length, bool const sst)
 {
+  char* endptr;
+  char* startptr;
   wsrep_uuid_t uuid;
   size_t const uuid_len= wsrep_uuid_scan(value, length, &uuid);
-  wsrep_seqno_t const seqno= strtoll(value + uuid_len + 1, NULL, 10);
+  startptr= (char *)value + uuid_len + 1;
+  wsrep_seqno_t const seqno= parse_value<uint64_t>(&startptr, &endptr);
+
+  if (*startptr == ',')
+  {
+    startptr++;
+    wsrep_gtid_server.domain_id= parse_value<uint32_t>(&startptr, &endptr);
+    startptr++;
+    wsrep_gtid_server.server_id= parse_value<uint32_t>(&startptr, &endptr);
+    startptr++;
+    wsrep_gtid_server.seqno(parse_value<uint64_t>(&startptr, &endptr));
+  }
 
   char start_pos_buf[FN_REFLEN];
   memcpy(start_pos_buf, value, length);
@@ -480,7 +508,7 @@ bool wsrep_provider_update (sys_var *self, THD* thd, enum_var_type type)
   if (!rcode)
     refresh_provider_options();
 
-  wsrep_set_wsrep_on();
+  wsrep_set_wsrep_on(thd);
   mysql_mutex_lock(&LOCK_global_system_variables);
 
   return rcode;
@@ -499,8 +527,8 @@ void wsrep_provider_init (const char* value)
   }
 
   if (wsrep_provider) my_free((void *)wsrep_provider);
-  wsrep_provider= my_strdup(value, MYF(0));
-  wsrep_set_wsrep_on();
+  wsrep_provider= my_strdup(PSI_INSTRUMENT_MEM, value, MYF(0));
+  wsrep_set_wsrep_on(NULL);
 }
 
 bool wsrep_provider_options_check(sys_var *self, THD* thd, set_var* var)
@@ -536,7 +564,7 @@ void wsrep_provider_options_init(const char* value)
 {
   if (wsrep_provider_options && wsrep_provider_options != value)
     my_free((void *)wsrep_provider_options);
-  wsrep_provider_options= (value) ? my_strdup(value, MYF(0)) : NULL;
+  wsrep_provider_options= value ? my_strdup(PSI_INSTRUMENT_MEM, value, MYF(0)) : NULL;
 }
 
 bool wsrep_reject_queries_update(sys_var *self, THD* thd, enum_var_type type)
@@ -580,6 +608,15 @@ bool wsrep_debug_update(sys_var *self, THD* thd, enum_var_type type)
   return false;
 }
 
+bool
+wsrep_gtid_seq_no_check(sys_var *self, THD *thd, set_var *var)
+{
+  ulonglong new_wsrep_gtid_seq_no= var->save_result.ulonglong_value;
+  if (wsrep_gtid_mode && new_wsrep_gtid_seq_no > wsrep_gtid_server.seqno())
+    return false;
+  return true;
+}
+
 static int wsrep_cluster_address_verify (const char* cluster_address_str)
 {
   /* There is no predefined address format, it depends on provider. */
@@ -619,7 +656,7 @@ bool wsrep_cluster_address_update (sys_var *self, THD* thd, enum_var_type type)
      connections. Closing clients may need to get LOCK_global_system_variables
      at least in MariaDB.
   */
-  char *tmp= my_strdup(wsrep_cluster_address, MYF(MY_WME));
+  char *tmp= my_strdup(PSI_INSTRUMENT_ME, wsrep_cluster_address, MYF(MY_WME));
   WSREP_DEBUG("wsrep_cluster_address_update: %s", wsrep_cluster_address);
   mysql_mutex_unlock(&LOCK_global_system_variables);
 
@@ -653,8 +690,8 @@ void wsrep_cluster_address_init (const char* value)
               (wsrep_cluster_address) ? wsrep_cluster_address : "null",
               (value) ? value : "null");
 
-  my_free((void*) wsrep_cluster_address);
-  wsrep_cluster_address= my_strdup(value ? value : "", MYF(0));
+  my_free(const_cast<char*>(wsrep_cluster_address));
+  wsrep_cluster_address= my_strdup(PSI_INSTRUMENT_MEM, safe_str(value), MYF(0));
 }
 
 /* wsrep_cluster_name cannot be NULL or an empty string. */
@@ -727,7 +764,7 @@ void wsrep_node_address_init (const char* value)
   if (wsrep_node_address && strcmp(wsrep_node_address, value))
     my_free ((void*)wsrep_node_address);
 
-  wsrep_node_address= (value) ? my_strdup(value, MYF(0)) : NULL;
+  wsrep_node_address= value ? my_strdup(PSI_INSTRUMENT_MEM, value, MYF(0)) : NULL;
 }
 
 static void wsrep_slave_count_change_update ()
@@ -929,6 +966,11 @@ bool wsrep_max_ws_size_update(sys_var *self, THD *thd, enum_var_type)
   return refresh_provider_options();
 }
 
+bool wsrep_mode_check(sys_var *self, THD* thd, set_var* var)
+{
+  return false;
+}
+
 #if UNUSED /* eaec266eb16c (Sergei Golubchik  2014-09-28) */
 static SHOW_VAR wsrep_status_vars[]=
 {
@@ -1037,6 +1079,11 @@ int wsrep_show_status (THD *thd, SHOW_VAR *var, void *,
     var->type= SHOW_ARRAY;
     var->value= (char *) &mysql_status_vars;
   }
+  else
+  {
+    var->type= SHOW_CHAR;
+    var->value= (char*) "0";
+  }
   return 0;
 }
 
@@ -1044,3 +1091,33 @@ void wsrep_free_status (THD* thd)
 {
   thd->wsrep_status_vars.clear();
 }
+
+bool wsrep_gtid_domain_id_update(sys_var* self, THD *thd, enum_var_type)
+{
+  WSREP_DEBUG("wsrep_gtid_domain_id_update: %llu",
+              wsrep_gtid_domain_id);
+  wsrep_gtid_server.domain_id= wsrep_gtid_domain_id;
+  return false;
+}
+
+bool wsrep_strict_ddl_update(sys_var *self, THD* thd, enum_var_type var_type)
+{
+  // In case user still sets wsrep_strict_ddl we set new
+  // option to wsrep_mode
+  if (wsrep_strict_ddl)
+    wsrep_mode|= WSREP_MODE_STRICT_REPLICATION;
+  else
+    wsrep_mode&= (~WSREP_MODE_STRICT_REPLICATION);
+  return false;
+}
+
+bool wsrep_replicate_myisam_update(sys_var *self, THD* thd, enum_var_type var_type)
+{
+  // In case user still sets wsrep_replicate_myisam we set new
+  // option to wsrep_mode
+  if (wsrep_replicate_myisam)
+    wsrep_mode|= WSREP_MODE_REPLICATE_MYISAM;
+  else
+    wsrep_mode&= (~WSREP_MODE_REPLICATE_MYISAM);
+  return false;
+}
