@@ -210,13 +210,7 @@ int maria_extra(MARIA_HA *info, enum ha_extra_function function,
 	    info->last_key.data + share->base.max_key_length*2,
 	    info->save_lastkey_data_length + info->save_lastkey_ref_length);
       info->update=	info->save_update | HA_STATE_WRITTEN;
-      if (info->lastinx != info->save_lastinx)             /* Index changed */
-      {
-        info->lastinx = info->save_lastinx;
-        info->last_key.keyinfo= info->s->keyinfo + info->lastinx;
-        info->last_key.flag= 0;
-        info->page_changed=1;
-      }
+      info->lastinx=	info->save_lastinx;
       info->cur_row.lastpos= info->save_lastpos;
       info->last_key.data_length= info->save_lastkey_data_length;
       info->last_key.ref_length= info->save_lastkey_ref_length;
@@ -235,6 +229,9 @@ int maria_extra(MARIA_HA *info, enum ha_extra_function function,
     info->lock_wait= MY_SHORT_WAIT;
     break;
   case HA_EXTRA_NO_KEYS:
+    if (share->s3_path)                    /* Not supported with S3 */
+      break;
+
     /* we're going to modify pieces of the state, stall Checkpoint */
     mysql_mutex_lock(&share->intern_lock);
     if (info->lock_type == F_UNLCK)
@@ -375,16 +372,16 @@ int maria_extra(MARIA_HA *info, enum ha_extra_function function,
       if (end_io_cache(&info->rec_cache))
         error= 1;
     }
-    if (share->kfile.file >= 0)
+    if (share->kfile.file >= 0 && share->s3_path == 0)
     {
       if (do_flush)
       {
         /* Save the state so that others can find it from disk. */
-        if ((share->changed &&
-             _ma_state_info_write(share,
+        if (share->changed &&
+            (_ma_state_info_write(share,
                                   MA_STATE_INFO_WRITE_DONT_MOVE_OFFSET |
-                                  MA_STATE_INFO_WRITE_FULL_INFO)) ||
-            mysql_file_sync(share->kfile.file, MYF(0)))
+                                  MA_STATE_INFO_WRITE_FULL_INFO) ||
+             mysql_file_sync(share->kfile.file, MYF(0))))
           error= my_errno;
       }
       else
@@ -396,7 +393,7 @@ int maria_extra(MARIA_HA *info, enum ha_extra_function function,
       }
     }
     if (share->data_file_type == BLOCK_RECORD &&
-        share->bitmap.file.file >= 0)
+        share->bitmap.file.file >= 0 && share->s3_path == 0)
     {
       DBUG_ASSERT(share->bitmap.non_flushable == 0 &&
                   share->bitmap.changed == 0);
